;;;; -*-  Mode: LISP; Syntax: Common-Lisp; Base: 10                          -*-
;;;; ---------------------------------------------------------------------------
;;;; File name: bldrules.lsp
;;;;    System: LTRE
;;;;   Version: 4.1
;;;;    Author: Ken Forbus, Johan de Kleer, John Everett
;;;;   Created: May 1, 1986
;;;;  Modified: everett on Fri Aug 2 9:19:35 1996
;;;;   Purpose: Create compilable rules for LTRE
;;;; Copyright: 1986, 1989, 1990, 1991, 1995 Kenneth D. Forbus, 
;;;;            Northwestern University, and Johan de Kleer, Xerox Corporation.
;;;;            All Rights Reserved.
;;;; ---------------------------------------------------------------------------

(in-package :LTRE)

;;; DEBUGGING RULES
;;; ----------------------------------------------------------------------------
;;; To have rule structs save extra information about their function name and
;;; args in a slot called rule-name, do the following:
;;; 1. Set *debug-ltre-rules* to T.
;;; 2. Reload ALL ltre files
;;; 3. RECOMPILE any rule files (including setrule) that you are currently using
;;; Rule files will be about 10-20% larger when compiled with this option.

(defparameter *debug-ltre-rules* NIL "When T causes rules to save extra info")

(if *debug-ltre-rules*
   (pushnew :DEBUG-LTRE-RULES *features*)
   (setf *features* (remove :DEBUG-LTRE-RULES *features*)))


;;; GLOBAL VARIABLE AND DATA STRUCTURE DEFINITIONS
;;; ----------------------------------------------------------------------------
;;; The variables below are accumulators used during rule generation.

(defvar *end-forms* nil "Cache of procedure definitions")
(defvar *rule-indexing* nil "Cache of rule indexing forms")
(defvar *matcher-table* nil "Lambda-bound table for reusing matchers")

;;; The variables below are lambda-bound during rule construction.  *bound-vars* 
;;; keeps track of pattern variables that have bindings from enclosing triggers.  
;;; Because the rule system now checks all nodes for the right trigger conditions 
;;; at the time of execution, *label-var-pairs* maintains a list of node vars
;;; and trigger conditions which is used to generate the lambda lists of inner rule 
;;; bodies and to generate the right label tests at compile time.  *Nesting-level* 
;;; is used to create node variables (eg n1, n2 n3) that correspond to the level of 
;;; the trigger condition that the variable refers to.  *gc-able-nodes* is a list 
;;; of the node variables that will be bound to gc-able nodes.  Because we must gc 
;;; any rule structs containing closures over such nodes, the matchers of all such 
;;; rule structs push the rule struct onto the tms-node-rules-to-gc slot of all 
;;; such nodes that enclose the matcher in their scopes.  In effect, this passes 
;;; upward a pointer to an inner rule, going against the normal flow of info.  
;;; *envs* caches all body-function lambda-lists so we can determine whether a 
;;; *gc-able-node* is going to be contained in a closure.  *current-rule* points to 
;;; the rule struct being executed, providing a run-time handle for code that must 
;;; be created at compile-time.

(defvar *bound-vars* nil "Current lexical environment")
(defvar *label-var-pairs* nil "List of trigger condition/node var pairs") 
(defvar *nesting-level* 0 "How many nodes scope current rule")
(defvar *gc-able-nodes* nil "List of trigger nodes that could be gc'd")
(defvar *envs* nil "List of all body-fn envs created")

(defvar *current-rule* nil "Bound to rule struct being executed")

(defvar *file-counter* 0 "Enable file-specific prefixes")
(defvar *file-prefix* "" "Enable file-specific prefixes")

#-DEBUG-LTRE-RULES
(defstruct (rule (:PRINT-FUNCTION rule-print-procedure))
   counter      ; Unique ID for easy lookup
   dbclass      ; Dbclass of associated pattern
   matcher      ; Procedure that performs the match
   body         ; Procedure that does the rules' work
   triggers     ; A list of label-node pairs
   thread       ; pointer for threading rules
   bindings     ; bindings from matcher for body to be applied to
   tms-nodes    ; nodes that rule refers to
   )

#+DEBUG-LTRE-RULES
(defstruct (rule (:PRINT-FUNCTION rule-print-procedure))
   counter      ; Unique ID for easy lookup
   name         ; helpful for debugging, a waste of space otherwise
   dbclass      ; Dbclass of associated pattern
   matcher      ; Procedure that performs the match
   body         ; Procedure that does the rules' work
   triggers     ; A list of label-node pairs
   thread       ; pointer for threading rules
   bindings     ; bindings from matcher for body to be applied to
   tms-nodes    ; nodes that rule refers to
   )

(defun rule-print-procedure (r st ignore)
   (declare (ignore ignore))
  (format st "<Rule ~D>" (rule-counter r)))

(defun rule-eq (r1 r2)
   (= (rule-counter r1) (rule-counter r2)))

(defmacro Rule-File (prefix)
   `(eval-when (compile load eval)
       (setq *file-counter* 0)
       (setq *file-prefix* ,prefix)))

(defun count-rules ()
   (let ((counter 0))
      (map-over-rules
         #'(lambda (r)
              (declare (ignore r))
              (incf counter)))
      counter))


;;; BUILDING AND INSTALLING RULES--TOPLEVEL FUNCTIONS AND MACROS
;;; ----------------------------------------------------------------------------
;;; These procedures arrange for rules nested in rules to be created as internal
;;; rules and for the top-level rule functions to be added to the dbclass table
;;; The real work of generating rules occurs in a mutual recursion between 
;;; build-rule and generate-rule-body, in the next section.

(defmacro rule (triggers &rest body) (do-rule triggers body))

(defun do-rule (triggers body)
   (check-trigger-vars triggers)
   (let ((*end-forms* nil)        ;; accumulates procedure definitions
         (*rule-indexing* nil)    ;; accumulates indexing forms
         (*ltre* *ltre*)          ;; the LTRE the rule will use
         (*nesting-level* 0)      ;; how many triggers enclose current work
         (*envs* nil)             ;; all body-fn envs created
         (*gc-able-nodes* nil)    ;; nodes that could be gcd
         (*label-var-pairs* nil)) ;; list of form ((:TRUE . n1) (:INTERN . n2))
      ;;Uncomment these if you have a serious rule expansion problem.
      ;;(format t "~% Adding rule with triggers: ") (pprint triggers)
      ;;(format t "~% And body: ") (pprint body)
      (add-rule (car triggers)
         (subst 'internal-rule 'rule
            (make-nested-rule (cdr triggers) body)))
      ;;Returning this ensures that all procedure definitions
      ;;are executed before any indexing occurs.
      `(progn ,@*end-forms* ,@*rule-indexing*)))

(defun check-trigger-vars (triggers &aux (tr-vars nil))
   "Ensures that each trigger has a unique :var variable"
   (every #'(lambda (tr)
               (let ((tr-var (cadr (member :var tr :test #'eq))))
                  (cond ((and tr-var (member tr-var tr-vars))
                         (error "Duplicated trigger :var ~A in ~A~%"
                            tr-var triggers))
                        (t (push tr-var tr-vars)))))
      triggers))

(defmacro internal-rule (triggers &rest body)
   "All but rule corresponding to outermost trigger are internal-rules"
   `(add-internal-rule  ,(car triggers)
       ,(make-nested-rule (cdr triggers) body)))

(defmacro add-internal-rule (trigger body)
   "Form to index this rule must appear in body of rule directly containing it"
   (build-rule trigger body))

(defun make-nested-rule (triggers body)
   (cond ((null triggers) body)
         (t `((add-internal-rule ,(car triggers)
                 ,(make-nested-rule (cdr triggers) body))))))

(defun add-rule (trigger body)
   "Executed after fn definition--nil in build-rule indicates top-level" 
   (push (build-rule trigger body nil) *rule-indexing*)
   nil)


;;; CREATING RULE FUNCTIONS--THE REAL WORK GETS DONE HERE
;;; ----------------------------------------------------------------------------
;;; Building a rule involves analyzing the trigger to see what special-purpose 
;;; unifier is needed, constructing a function to do the work of the body, and
;;; building a form to index it.

(defmacro with-pushed-trigger-conditions (condition pattern &body body)
   "Lambda-bind caches of current level, trigger-conditions, and node vars"
   `(let* ((*nesting-level* (1+ *nesting-level*))
           (node-var (intern (format nil "N~A" *nesting-level*)))
           (*label-var-pairs* (cons (cons ,condition node-var)
                                 *label-var-pairs*))
           (*gc-able-nodes* (if (funcall (ltre-gc-test *ltre*) pattern)
                               (push node-var *gc-able-nodes*)
                               *gc-able-nodes*)))
       ,@body))

(defun build-rule (trigger body &optional (nested? T) 
                                &aux match-procedure body-procedure)
   (multiple-value-bind (pattern condition var test)
       (parse-rule-trigger trigger)
      (with-pushed-trigger-conditions condition pattern
       (setq match-procedure ;;reuse matchers whenever possible to save space
          (procure-match-procedure pattern var test))
       (setq body-procedure (generate-body-procedure pattern var body))
       (push body-procedure *end-forms*)
       `(insert-rule ;;return form to index rule
           (get-dbclass ,(get-trigger-dbclass pattern))
           #+DEBUG-LTRE-RULES ',(make-rule-name-for-debugging body-procedure)
           (function ;;the match function for rule
              (lambda (p)
                 (,(cadr match-procedure) ,@(make-matcher-lambda-list))))
           (function ;;the body function for rule
              ,(cond ((or *bound-vars* 
                          (and nested? (not (eq condition :INTERN))))
                      (let ((tv (pattern-free-variables trigger)))
                         `(lambda ,(cons (current-node-var) tv)
                             ,(make-closure-body body-procedure tv))))
                     ;; When nested? is nil we're at the top level and shouldn't
                     ;; bother returning a closure
                     (T (cadr body-procedure))))))))

(defun procure-match-procedure (pattern var test &aux existing)
   "Either return an already-created matcher or make a new one as needed"
   (let ((matcher (generate-match-procedure pattern var test)))
      (cond ((and *matcher-table*
                  (setq existing (gethash (cddr matcher) *matcher-table*)))
             existing)
            (*matcher-table*
               (setf (gethash (cddr matcher) *matcher-table*) matcher)
               (push matcher *end-forms*)
               matcher)
            (t (push matcher *end-forms*)
               matcher))))

(defun current-node-var ()
   (cdar *label-var-pairs*))

(defun make-rule-name-for-debugging (body-procedure)
   "Make a rule-struct name to facilitate debugging of rules"
   (let ((env (third body-procedure)))
      `(,(cadr body-procedure)
        ,(count-if-not #'variable? env)
        ,(remove-if-not #'variable? env))))

(defun parse-rule-trigger (trigger)
   ;; Trigger syntax is (<condition> <pattern> .  <options>)
   ;; where <condition> =  :INTERN | :TRUE | :FALSE
   ;; and <options> can be empty, or 
   ;; :TEST <code> and/or :VAR <var>, where <code> must be
   ;; non-nil for the match to succeed, and <var> will be
   ;; bound to the whole pattern.
   (when (oddp (length trigger)) ;;syntax check
      (error "Malformed trigger: ~A~%" trigger))
   (values (cadr trigger)
      (cond ((member (car trigger) '(:INTERN :TRUE :FALSE)) (car trigger))
            (t (error "~% Unknown belief condition ~A in trigger ~A."
                  (car trigger) trigger)))
      (cadr (member :VAR (cddr trigger)))
      (cadr (member :TEST (cddr trigger)))))

(defun make-closure-body (fn-body lambda-env)
   "Package up vars in current scope that won't be available later"
   `(,(cadr fn-body)                          ;;function name
     ,@(mapcar #'cdr *label-var-pairs*)       ;;node variables
     ,@lambda-env                             ;;vars passed into lambda
     ,@(scratchout lambda-env *bound-vars*))) ;;vars closed over
    
(defun get-trigger-dbclass (trigger)
  (cond ((variable? trigger)
	 (if (member trigger *bound-vars*)  trigger
	     (error "~%Trigger dbclass is unbound -- ~A." trigger)))
        ((atom trigger) (list 'QUOTE trigger))
        (t (get-trigger-dbclass (car trigger)))))


;;; GENERATING THE BODY FUNCTION
;;; ----------------------------------------------------------------------------
;;; Macros for creating the correct local environment at each level of recursion

(defmacro with-pushed-variable-bindings (new-bindings &body body)
   "Lambda-bind *bound-vars* to correct bindings for current recursion level"
   `(let ((*bound-vars* (append ,new-bindings
                           (scratchout ,new-bindings
                              *bound-vars*))))
       ,@body))

(defmacro with-pushed-envs (newly-bound &body body)
   "Save envs in case we have to cache on an outer node"
   `(let ((*envs* (cons (make-env-list ,newly-bound) *envs*)))
       ,@body))

(defun make-env-list (newly-bound)
   `(,@(mapcar #'cdr *label-var-pairs*) ;;node variables
     ,@(append newly-bound
          (scratchout newly-bound *bound-vars*))))

(defun generate-body-procedure (pattern var body &aux newly-bound env fname)
   (setq newly-bound (pattern-free-variables pattern))
    (when var (push var newly-bound))
   (with-pushed-envs newly-bound
      (setq env (car *envs*))
      (setq body (with-pushed-variable-bindings newly-bound 
                 (fully-expand-body body)))
      (setq fname (generate-rule-procedure-name pattern))
      `(defun ,fname ,env
          ,@(cond ((every 'intern? *label-var-pairs*) ;; just do it
                   `,body)
                  (t ;;Must check if this and all prior nodes' belief state
                     ;;match the rule's trigger conditions
                     `((cond ,@(make-node-trigger-caching-tests)
                             ;;If all tests passed, then execute body
                             (T ,@body))))))))

(defun make-node-trigger-caching-tests ()
   "Return tests for all nodes up to and including current node"
   (mapcan 
      #'(lambda (pair)
           (ecase (car pair)
              (:TRUE
               `(((not (true-node? ,(cdr pair)))
                  (cache-rule-on-node ,(cdr pair) ,(car pair)))))
              (:FALSE
               `(((not (false-node? ,(cdr pair)))
                  (cache-rule-on-node ,(cdr pair) ,(car pair)))))
              (:INTERN NIL)))
      *label-var-pairs*))

(defun intern? (label-var-pair)
   (eq (car label-var-pair) :INTERN))


;;; GENERATING THE OPEN-CODED MATCHING PROCEDURE
;;; ----------------------------------------------------------------------------
;;; Rule structs contain matcher and body closures.  If a gc'd node is within 
;;; one of these closures, we would like to delete that rule struct, as only bad 
;;; things can come of running it, and it's taking up space.  However, we need 
;;; a pointer to that rule struct.  The rule building procedures keep track of 
;;; all gc-able trigger nodes, and if a matcher body is in a lexical scope that 
;;; contains gc-able nodes (remember, we pass down all nodes in order to check 
;;; labels of all triggers prior to execution), then that matcher will, on 
;;; matching, push *current-rule* onto the rule-cache of each gc-able node.  
;;; Build-rule now generates matcher closures that contain these nodes.

;;; The reason for the function below is to ensure that we don't create 
;;; pattern-matching closures over the node of the fact that the matcher is 
;;; trying to match against--obviously, this node hasn't yet entered the 
;;; lexical scope of a rule body procedure, so it isn't present to be closed 
;;; over.  We cache this node instead in try-rule-on, because we do have direct 
;;; access to it there, via the datum that we just successfully matched 
;;; against.

(defun generate-match-procedure (pattern var test &aux proc)
   "Return two values, OK? and bindings when OK"
   (multiple-value-bind (tests binding-specs)
       ;;make special tests to check for this pattern
       (generate-match-body pattern (pattern-free-variables pattern)
          (if var (list (cons var 'P))) test)
      `(defun ,(generate-rule-procedure-name pattern) ,(make-matcher-lambda-list)
          ,@(make-gc-able-node-rule-caching-forms)
          (when (and ,@tests)
             (values T 
                (list ,@(if var '(P)) ,@(reverse binding-specs))
                ;; a keylist of node trigger-label.  By reversing order
                ;; of nodes and labels we can simply push node we matched on
                ;; onto this list at runtime and save a bunch of consing
                (list ,(caar *label-var-pairs*)
                   ,@(mapcan #'(lambda (pair)
                                  ;;note reversal of node & label
                                  (list (cdr pair) (car pair)))
                        (cdr *label-var-pairs*))))))))

(defun make-matcher-lambda-list ()
   (let* ((all-nodes (mapcar #'cdr (cdr *label-var-pairs*)))
          (gc-able-nodes (gc-able-nodes-in-lexical-scope))
          (other-nodes (scratchout gc-able-nodes all-nodes)))
      ;;first arg, P, is the pattern
      `(P ,@gc-able-nodes ,@*bound-vars* ,@other-nodes)))

(defun gc-able-nodes-in-lexical-scope ()
   "Return those gc-able nodes that matcher can close over"
   (intersection *gc-able-nodes* (car *envs*)))

(defun make-gc-able-node-rule-caching-forms ()
   "When any incoming node-var is gc-able, cache current rule on that node
    to enable later gc of rules that contain gc'd nodes in closures"
   (mapcar #'(lambda (node)
                `(pushnew *current-rule* (tms-node-rules-to-gc ,node)))
      (gc-able-nodes-in-lexical-scope)))

(defun scratchout (things-to-remove list-to-remove-from &aux result)
   "Preserves order of list-to-remove-from"
   (setq result (copy-list list-to-remove-from))
   (dolist (element things-to-remove result)
      (setq result (delete element (the list result)))))

(defun generate-rule-procedure-name (pattern)
   (intern (format nil "~A-~A-~A" 
              *file-prefix* pattern (incf *file-counter*))))


;;; RECURSIVE MACROEXPANSION
;;; ----------------------------------------------------------------------------
(defvar *macros-to-expand*
   '(rule internal-rule add-internal-rule with-pushed-variable-bindings
       with-pushed-trigger-conditions with-pushed-envs
       rlet rassert! rretract! assuming with-LTRE))

(defun fully-expand-body (body)
  ;; New version -- only expands when it is clear that expansion is needed.
  (cond ((null body) nil)
	((not (listp body)) body)
        ((symbolp (car body))
         (cond ((member (car body) *macros-to-expand*)
                (when (eql (car body) 'with-pushed-variable-bindings)
                   (format *Debug-io*
                    "Expanding a with-pushed-variable-bindings~%"))
		(fully-expand-body (macroexpand body)))
	       (t (cons (car body)
			(fully-expand-body (cdr body))))))
	(t (cons (fully-expand-body (car body))
		 (fully-expand-body (cdr body))))))


;;; COMPILING FILES OF RULES
;;; ----------------------------------------------------------------------------
;;; The *matcher-table* enables one to reuse identical matcher bodies, resulting 
;;; in a reduction in compiled rulebase size of up to 33%.

(defun load-rule-files (rule-path filenames 
                          &key (gc-test #'default-gc-p)
                               (outpath rule-path) 
                               (reuse-matchers? T)
                               (verbose nil)
                               (action :compile-if-newer))
   (unwind-protect 
        (let (#+MCL(ccl::*suppress-compiler-warnings* t)
              (*matcher-table* 
                 (when reuse-matchers?
                    (make-hash-table :size 500 :test #'equal))))
           (unless *ltre* 
              (in-ltre (create-ltre "Compiler Scratchpad"
                          :setrule nil
                          :gc-test gc-test)))
           (load-files rule-path filenames
              :action action
              :verbose verbose
              :outpath outpath))
      ;;make darn sure that *matcher-table* is reset
      (setq *matcher-table* nil))
   (beep *terminal-io*)) ;;signal completion

;;; ----------------------------------------------------------------------------
;;; END OF CODE