;;;; -*-  Mode: LISP; Syntax: Common-Lisp; Base: 10                          -*-
;;;; ---------------------------------------------------------------------------
;;;; File name: contrad.lsp
;;;;    System: LTRE
;;;;   Version: 4.1 
;;;;    Author: Ken Forbus and Johan de Kleer
;;;;   Created: December 21, 1995
;;;;  Modified: everett on Sat Jul 6 13:44:22 1996
;;;;   Purpose: Handle LTMS contradictions
;;;; Copyright: 1986, 1989, 1990, 1991, 1995 Kenneth D. Forbus, 
;;;;            Northwestern University, and Johan de Kleer, Xerox Corporation.
;;;;            All Rights Reserved.
;;;; ---------------------------------------------------------------------------

(in-package :LTRE)


;;; CONTRADICTION HANDLING INTERFACE
;;; ----------------------------------------------------------------------------
(defun check-for-contradictions (ltms &aux violated-clauses)
  (setq violated-clauses
	(delete-if-not #'(lambda (c) (violated-clause? c))
		       (ltms-violated-clauses ltms)))
  ;; Update cache now because handlers can throw 
  (setf (ltms-violated-clauses ltms) violated-clauses)
  (if violated-clauses (contradiction-handler ltms violated-clauses)))

(defun is-contradictory? (ltms &aux violated-clauses)
   (setf (ltms-violated-clauses ltms)
         (delete-if-not #'(lambda (c) (violated-clause? c))
            (ltms-violated-clauses ltms)))
   (setf (ltms-pending-contradictions ltms)
         (delete-if-not #'(lambda (c) (violated-clause? c))
            (ltms-pending-contradictions ltms)))
   ;;Remove-duplicates redresses sins of append
   ;;Unfortunately, nconc is dangerous in this situation
   (remove-duplicates
      (append (ltms-violated-clauses ltms)
         (ltms-pending-contradictions ltms))
      :test #'clause-eq))

(defun contradiction-handler (ltms violated-clauses)
  (cond ((not (ltms-checking-contradictions ltms))
	 ;;Save contra info for later handling
         (setf (ltms-pending-contradictions ltms)
               (delete-if-not #'(lambda (c) (violated-clause? c))
                  (ltms-pending-contradictions ltms)))
         (dolist (violated-clause violated-clauses) 
            ;;Better safe than sorry.
            (pushnew violated-clause (ltms-pending-contradictions ltms))))
        (t (dbg-ltre :CONTRA "~%Entering contradiction handler in ~A on ~A."
            ltms violated-clauses)
           (dolist (handler (ltms-contradiction-handlers ltms))
              (dbg-ltre :CONTRA "~%    Trying ~A.." handler)
              (if (funcall handler violated-clauses ltms) (return T))))))

(defmacro without-contradiction-check (ltms &body body)
  (contradiction-check ltms nil body))

(defmacro with-contradiction-check (ltms &body body)
  (contradiction-check ltms t body))

(defun contradiction-check (ltms flag body)
  `(let* ((.ltms. ,ltms)
	  (.old-value. (ltms-checking-contradictions .ltms.)))
     (unwind-protect
	 (progn (setf (ltms-checking-contradictions .ltms.) ,flag)
		,@body)
       (setf (ltms-checking-contradictions .ltms.) .old-value.))))

(defmacro with-contradiction-handler (ltms handler &body body)
   `(let ((.ltms. ,ltms))
       (unwind-protect
            (progn (push ,handler (ltms-contradiction-handlers .ltms.))
               ,@body)
          (pop (ltms-contradiction-handlers .ltms.)))))

(defmacro with-assumptions (assumption-values &body body)
   ;;Allows assumptions to be made safely, and retracted properly
   ;;even if non-local exits occur.
   `(unwind-protect (progn (dolist (av ,assumption-values)
                              (enable-assumption (car av) (cdr av)))
                       ,@body)
       (dolist (av ,assumption-values) (retract-assumption (car av)))))


;;; INQUIRING ABOUT WELL-FOUNDED SUPPORT
;;; ----------------------------------------------------------------------------
(defun support-for-node (node &aux result support)
   (cond ((null (setq support (tms-node-support node))) nil)
         ((enabled-assumption? node) :ENABLED-ASSUMPTION)
         (t (dolist (pair (clause-literals support))
               (unless (eq (car pair) node)
                  (push node result)))
            (values result (clause-informant support)))))

(defun assumptions-of-node (node)
   (cond ((enabled-assumption? node) (list node))
         ((known-node? node)
          (assumptions-of-clause (tms-node-support node)))))

(defun assumptions-of-clause (in-clause &aux)
   (when (clause-p in-clause)
      (do ((clause-queue (list in-clause)
            (nconc (cdr clause-queue) new-clauses))
           (mark (list nil))
           (node nil)
           (new-clauses nil nil)
           (assumptions nil))
          ((null clause-queue) assumptions)
         (dolist (term-pair (clause-literals (car clause-queue)))
            (setq node (car term-pair))
            (unless (eq (tms-node-mark node) mark)
               (unless (eq (tms-node-label node) (cdr term-pair))
                  (cond ((enabled-assumption? node)
                         (push node assumptions))
                        ((null (tms-node-support node))
                         (ltms-error "node is unknown" node))
                        (t (push (tms-node-support node) new-clauses))))
               (setf (tms-node-mark node) mark))))))


;;; A SIMPLE USER INTERFACE
;;; ----------------------------------------------------------------------------
(eval-when (eval compile load)
  (proclaim '(special *contra-assumptions*
	    *contra-clause*)))

(defun ask-user-handler (contradictions ltms)
  (declare (ignore ltms))
  (dolist (contradiction contradictions)
    (if (violated-clause? contradiction)
	(handle-one-contradiction contradiction))))

(defun handle-one-contradiction (violated-clause)
  (let ((*contra-assumptions* (assumptions-of-clause violated-clause))
	(the-answer nil))
    (unless *contra-assumptions* (ltms-error "Global contradiction"
					     violated-clause))
    (setq *contra-clause* violated-clause)
    (format t "~%Contradiction found:")
    (print-contra-list *contra-assumptions*)
    (format t "~%Call (TMS-ANSWER <number>) to retract assumption.")
    (setq the-answer
	  (catch 'tms-contradiction-handler
	    (cerror "Continue LTRE processing (retract an assumption first!)"
		    "LTMS contradiction found")))
    (if the-answer
	(retract-assumption (nth (1- the-answer)
				 *contra-assumptions*)))))

(defun print-contra-list (nodes)
  (do ((counter 1 (1+ counter))
       (nn nodes (cdr nn)))
      ((null nn))
    (format t "~%~A ~A" counter
	    (node-string (car nn)))))

(defun tms-answer (num)
  (if (integerp num)
      (if (> num 0)
	  (if (not (> num (length *contra-assumptions*)))
	      (throw 'tms-contradiction-handler num)
	      (format t "~%Ignoring answer, too big."))
	  (format t "~%Ignoring answer, too small"))
      (format t "~%Ignoring answer, must be an integer.")))

(defun avoid-all (contradictions ignore &aux culprits culprit sign)
  (declare (ignore ignore))
  (dolist (contradiction contradictions)
    (when (violated-clause? contradiction)
      (unless (setq culprits (assumptions-of-clause contradiction))
	(ltms-error "Total contradiction" contradiction))
      (setq culprit (car culprits)
	    sign (tms-node-label culprit))
      (retract-assumption culprit)
      (add-nogood culprit sign culprits)
      t)))

(defun clause-antecedents (clause &aux result)
  (dolist (pair (clause-literals clause) result)
    (unless (eq (tms-node-support (car pair)) clause)
      (push (car pair) result))))

(defun signed-node-string (node)
  (if (true-node? node) (node-string node)
      (format nil "~:[Unknown~;Not~][~A]"
	      (false-node? node) (node-string node))))

(defun node-consequences (node &aux conseq conseqs)
  (dolist (cl (ecase (tms-node-label node)
		(:TRUE (tms-node-false-clauses node))
		(:FALSE (tms-node-true-clauses node))))
    (unless (eq cl (tms-node-support node))
      (setq conseq (clause-consequent cl))
      (if conseq (push conseq conseqs))))
  conseqs)

(defun why-node (node &optional (simplify-informant? T) (stream t))
   (cond ((unknown-node? node)
          (format stream "~%~A is unknown." (node-string node))
          nil)
         ((enabled-assumption? node)
          (format stream "~%~A is ~A <~A>"
             (node-string node)
             (tms-node-label node) (tms-node-support node))
          nil)
         (t (format stream "~%~A is ~A via ~A on"
               (node-string node)
               (tms-node-label node)
               (or (if (clause-informant (tms-node-support node))
                      (let ((inf (clause-informant 
                                  (tms-node-support node))))
                         (if simplify-informant? (compact-informant inf) inf))
                      (tms-node-support node))))
            (dolist (term-pair (clause-literals (tms-node-support node)))
               (unless (equal (tms-node-label (car term-pair))
                          (cdr term-pair))
                  (format stream "~%   ~A is ~A"
                     (node-string (car term-pair))
                     (tms-node-label (car term-pair)))))))
   node)

(defun compact-informant (inf)
  (cond ((listp inf) (third inf))
	(t inf)))

(defun why-nodes (ltms)
  (maphash #'(lambda (ignore n)
	       (declare (ignore ignore))
	       (why-node n)) (ltms-nodes ltms)))

(defun pretty-print-clauses (ltms)
  (walk-clauses ltms #'(lambda (l)
			 (format T "~% ")
			 (pretty-print-clause l))))

(defun pretty-print-clause (clause)
  (format T "(:OR")
  (dolist (literal (clause-literals clause))
    (format T " ~:[(:NOT ~A)~;~A~]"
	    (eq :TRUE (cdr literal)) (node-string (car literal))))
  (format T ")"))

(defun show-node-consequences (node)
  (let ((conseqs (node-consequences node)))
    (cond (conseqs 
	   (format t "~% Consequences of ~A:" (signed-node-string node))
	   (dolist (conseq conseqs)
		   (format t "~%  ~A" (signed-node-string conseq))))
	  (t (format t "~% ~A has no consequences." (node-string node))))))

(defun node-show-clauses (node)
  (format t "For ~A:" (node-string node))
  (dolist (cl (tms-node-true-clauses node))
    (format T "~%") (pretty-print-clause cl))
  (dolist (cl (tms-node-false-clauses node))
    (format T "~%") (pretty-print-clause cl)))

(defun explore-network (node)
  (unless (known-node? node)
	  (format t "~% Sorry, ~A not believed." (node-string node))
	  (return-from explore-network node))
  (do ((queue nil)
       (current node)
       (mode :ante)
       (options nil)
       (olen 0)
       (done? nil))
      (done? current)
      (cond ((eq mode :ante)
	     (why-node current t)
	     (setq options (if (typep (tms-node-support current) 'clause)
			       (reverse 
                                (clause-antecedents
                                 (tms-node-support current)))
			     nil)))
	    (t ;; Looking at consequences
	     (show-node-consequences current)
	     (setq options (node-consequences current))))
      (setq olen (length options))
      (do ((good? nil)
	   (choice 0))
	  (good? (case good?
		       ((q cl-user::q :q) (return-from explore-network current))
		       ((c cl-user::c :c) (setq mode :conseq))
		       ((a cl-user::a :a) (setq mode :ante))
		       (0 (if queue
			      (setq current (pop queue))
			      (return-from explore-network current)))
		       (t (push current queue)
			  (setq current (nth (1- good?) options)))))
	  (format t "~%>>>")
	  (setq choice (read))
	  (if (or (member choice '(q cl-user::q :q))
		  (member choice '(c cl-user::c :c))
		  (member choice '(a cl-user::a :a))
		  (and (integerp choice)
		       (not (> choice olen))
		       (not (< choice 0))))
	      (setq good? choice)
	      (format t "~% Must be q, a, c or an integer from 0 to ~D."
		        olen)))))

;;; ----------------------------------------------------------------------------
;;; END OF CODE