;;;; -*-  Mode: LISP; Syntax: Common-Lisp; Base: 10                          -*-
;;;; ---------------------------------------------------------------------------
;;;; File name: defsys.lsp
;;;;    System: LTRE
;;;;   Version: 4.1 
;;;;    Author: Ken Forbus and Johan de Kleer
;;;;   Created: Dec 1 1993
;;;;  Modified: everett on Fri Aug 2 10:05:19 1996
;;;;   Purpose: Define LTRE environment, load functions
;;;; Copyright: 1986, 1989, 1990, 1991, 1995 Kenneth D. Forbus, 
;;;;            Northwestern University, and Johan de Kleer, Xerox Corporation.
;;;;            All Rights Reserved.
;;;; ---------------------------------------------------------------------------


(in-package :COMMON-LISP-USER)


;;; SETTING UP PACKAGES
;;; ----------------------------------------------------------------------------
;; Yes, it is obnoxious to unintern.  But we need rlet ourselves!
#+MCL (unexport 'RLET :CCL)

#+LUCID (defPackage :LTRE (:use :COMMON-LISP :CLOS))
  #+MCL (defPackage :LTRE (:use :COMMON-LISP))
#+ACLPC (defPackage :LTRE (:use :COMMON-LISP :COMMON-GRAPHICS))


;;; We follow a convention of placing all LTRE data in the CL-USER package.  
;;; This enables one to have several different problem-solvers work 
;;; concurrently on the same set of data.  However, it is possible that one 
;;; might want to isolate this data in its own package, so we have provided for 
;;; that by giving the CL-USER package a DATA nickname.  Should you want to put 
;;; your data in its own package, simply create a DATA package and comment out 
;;; the following.

(rename-package (find-package :CL-USER)
		"COMMON-LISP-USER"  ;keep name the same
		'("CL-USER" "USER"  ;replace standard nicknames
		  "DATA"))          ;add a new nickname


(in-package :LTRE)


;;; LOADING FACILITY
;;; ----------------------------------------------------------------------------
(defvar *root-path*    nil "Path stub to which all systems add their pathnames")
(defvar *path-sepchar* nil "Path separation character (system-dependent)")
(defvar *src-ext*      nil "Source extension for files")
(defvar *bin-ext*      nil "Binary or fasload extension for files")


;;; USER-SPECIFIC CONFIGURATION SECTION
;;; ============================================================================
;;; You must ensure that the following numbered items are set correctly for your
;;; particular hardware and software.


;1. PLATFORM-SPECIFIC SETTINGS -------------------------------------------------

#+RIOS (pushnew :MULTITASKING *features*)

;;; keeps CLIM from sticking commands in menubar
#+MCL (setq clim::*commands-in-menubar* nil) 


;2. SETTING THE LOCAL PATH STUB ------------------------------------------------
;;; This path stub must include a terminal path separation character.

(setq *root-path* "C:\\qrg\\")

      
;;; END OF USER-SPECIFIC CONFIGURATION
;;; ============================================================================
      


;;; HARDWARE-SPECIFIC SETTINGS 
;;; ----------------------------------------------------------------------------
;;; You should not change the following settings, except to add new cases.

(setq *path-sepchar*
      #+MAC ":"
      #+UNIX "/"
      #+ACLPC "\\" ;;the first backslash escapes
      )

(setq *src-ext*
  #+(OR MCL RIOS IRIX SUN)".lisp"
  #+ACLPC ".lsp"
  )

(setq *bin-ext*
  #+MCL ".fasl"
  #+ALLEGRO ".fasl"
  #+LUCID ".rbin"
  #+IRIX ".mbin"
  #+ACLPC ".fsl"
  )


;;; THE LOADER
;;; ----------------------------------------------------------------------------
;;; 1. :COMPILE-IF-NEWER results in an error (LUCID/RS6000) if the file
;;; contains, in this order, (1) a macro, (2) defun forms defining functions
;;; called by the macro, and (3) a top-level call to the macro, apparently
;;; because the compilation process does not incrementally load the forms of the
;;; file, so the first defun is undefined at the time that the compiler hits the
;;; top-level call.  A solution to this would be to have the :COMPILE-IF-NEWER
;;; first load the source, then compile and load the binary, but this would be
;;; slower.  Another solution would be to ensure that toplevel calls occur only
;;; in their own files, separate from their defining functions.  As of 2/25/95,
;;; going with this second "solution," but will implement the first if demand
;;; warrants.

(defmacro with-compiler-chatter (messages? &rest body)
  "Implementation-specific means of controlling compiler messaging"
  `(cond ((member :MCL *features*)
          (let (#+MCL(ccl::*suppress-compiler-warnings* ,messages?))
            ,@body))
         ((member :LUCID *features*)
          (cl-user::with-compiler-options (:warnings ,messages?
                                           :messages ,messages?)
            ,@body))
	 ((member :ALLEGRO *features*)
	  (let (#+ALLEGRO(lisp:*compile-verbose* ,messages?)
		#+ALLEGRO(lisp:*compile-print* ,messages?))
	    ,@body))
         ((member :ACLPC *features*)
          (handler-bind ((warning 'muffle-warning))
             ,@body))
         (t ;;Add branches for other implementation defintions
	  (progn ,@body))))

(defun load-files (path file-list
                   &KEY (action :COMPILE-IF-NEWER) (verbose t)
                   (outpath path)
                   (src-ext *src-ext*) (bin-ext *bin-ext*))
   (ecase action
      ((:COMPILE :COMPILE-IF-NEWER :SOURCE-IF-NEWER :LOAD-SOURCE :LOAD-BINARY)
       (dolist (file file-list)
          (load-file path file
             :action action
             :verbose verbose
             :outpath outpath
             :src-ext src-ext
             :bin-ext bin-ext)))
      (:LOAD-COMPILE
       (load-files path file-list
          :action :LOAD-SOURCE
          :verbose verbose
          :outpath outpath
          :src-ext src-ext
          :bin-ext bin-ext)
       (dolist (file file-list)
          (load-file path file
             :action :COMPILE
             :verbose verbose
             :outpath outpath
             :src-ext src-ext
             :bin-ext bin-ext)))))

(defun load-file (path file 
		  &KEY (action :COMPILE-IF-NEWER) (verbose t)
                  (outpath path)
		  (src-ext *src-ext*) (bin-ext *bin-ext*))
   (with-compiler-chatter verbose
    (ecase action
       (:COMPILE-IF-NEWER
        (let* ((src (make-full-file-spec path file src-ext))
               (bin (make-full-file-spec outpath file bin-ext))
               (src-exists? (probe-file src))
               (bin-exists? (probe-file bin)))
           (cond ((and src-exists? bin-exists?)
                  (cond ((>= (file-write-date bin)
                             (file-write-date src))
                         (load bin :verbose verbose))
                        (t (compile-file src :output-file bin)
                           (load bin :verbose verbose))))
                 (bin-exists? (load bin :verbose verbose))
                 (src-exists?
                  (compile-file src :output-file bin)
                  (load bin :verbose verbose))
                 (t (error "~%Can't find ~A or ~A~%" src bin)))))
       (:SOURCE-IF-NEWER
        (let* ((src (make-full-file-spec path file src-ext))
               (bin (make-full-file-spec outpath file bin-ext))
               (src-exists? (probe-file src))
               (bin-exists? (probe-file bin)))
           (cond ((and src-exists? bin-exists?)
                  (cond ((>= (file-write-date bin) (file-write-date src))
                         (load bin :verbose verbose))
                        (t (load src :verbose verbose))))
                 (bin-exists? (load bin :verbose verbose))
                 (src-exists? (load src :verbose verbose))
                 (t (error "~%Can't find ~A or ~A~%" src bin)))))
       (:COMPILE
        (compile-file
           (make-full-file-spec path file src-ext)
           :output-file (make-full-file-spec outpath file bin-ext)))
       (:LOAD-SOURCE
        (load (make-full-file-spec path file src-ext)
           :verbose verbose))
       (:LOAD-BINARY
        (load (make-full-file-spec path file bin-ext)
           :verbose verbose)))))

(defun make-full-file-spec (path file ext)
  "Ensures that a path separator character appears between path and file"
  (cond ((= (position *path-sepchar* (reverse path) :test #'string=) 0)
	 (concatenate 'string path file ext))
	(t (concatenate 'string path *path-sepchar* file ext))))

(defun make-source-file-name (path fname)
  (make-full-file-spec path fname *src-ext*))

(defun make-binary-file-name (path fname)
  (make-full-file-spec path fname *bin-ext*))

(defun make-file-name (path fname)
  (make-full-file-spec path fname ""))

(defun make-path (&rest strings)
  (concatenate 'string *root-path*
	       (reduce #'(lambda (str1 str2)
			   (concatenate 'string str1 *path-sepchar* str2))
		       strings)
	       *path-sepchar*))


;;; LTRE SYSTEM DEFINITIONS
;;; ----------------------------------------------------------------------------
(defparameter *ltre-version* "ltre-v41")

(defparameter *ltre-path* (make-path "ltre" *ltre-version*))

;;; The file setrule.lisp should also be compiled for efficiency.
;;; It can only be compiled after an LTRE has been created.

(defvar *set-rule-file* "setrule")
 
(load-file *ltre-path* "export" :action :load-source :verbose nil)

(defvar *ltre-files* 
   '("linter"    ;; Interface
     "gc-preds"  ;; Application-specific gc predicates
     "ltms"      ;; LTMS definitions
     "boolprop"  ;; Boolean constraint propagation
     "contrad"   ;; LTMS contradiction handling machinery
     "ldata"     ;; Database
     "bldrules"  ;; Rule creation functions
     "runrules"  ;; running, enqueuing, and gcing rules
     "unify"     ;; Unifier
     "funify"    ;; Open-coding of unification
     "cwa"       ;; Closed-world assumption mechanism
     "dds"       ;; Dependency-directed search facility
     "lcache"    ;; Caching system (allows easy KB reloading)
  ;; "laccept"   ;; acceptance tests for ltre, good place to start
  ;; "ltrecgex"  ;; ACLPC Common Graphics extensions to LTRE (not essential)
  ;; "linspect"  ;; ACLPC Common Graphics LTRE inspector (not essential)
     ))

(eval-when (eval load compile)
   (proclaim '(special *LTRE*)))

(defun load-ltre (&key (action :compile-if-newer)
                    (verbose t))
   (load-files *ltre-path* *ltre-files*
    :action action :verbose verbose))

(defun compile-ltre () 
   (load-files *ltre-path* *ltre-files*
    :action :COMPILE :verbose nil)
  (unless (and (boundp '*LTRE*)
               (not (null *LTRE*)))
     (create-ltre "DUMMY"))
   (load-files *ltre-path* (list *set-rule-file*)
    :action :COMPILE :verbose nil))

;;; ----------------------------------------------------------------------------
;;; END OF CODE
