;;;; -*- Mode: LISP; Syntax: Common-Lisp; Base: 10;                             -*-
;;;; ------------------------------------------------------------------------------
;;;; File name: linspect.lsp
;;;;    System: LTRE
;;;;   Version: 4.1
;;;;    Author: Kenneth D. Forbus, John Everett
;;;;   Created: Feb 7, 1996
;;;;  Modified: everett on Sat Jul 6 20:41:40 1996
;;;;   Purpose: GUI inspector for LTRE
;;;; Copyright: 1986, 1989, 1990, 1991, 1995 Kenneth D. Forbus, 
;;;;            Northwestern University, and Johan de Kleer, Xerox Corporation.
;;;;            All Rights Reserved.
;;;; ------------------------------------------------------------------------------

(in-package :LTRE)

;;; GENERAL
;;; ----------------------------------------------------------------------------
;;; The model here is that each hypertext window is a small, self-contained 
;;; piece of explanation/information.  Clicking on a link, if it adds to the 
;;; hypertext, creates a brand new window.  Kind of cluttered, to be sure, but 
;;; just right for building quickie interfaces for development purposes.

;;; (as-hypertext <stuff>) displays all of the information printed to 
;;; *standard-output* as a hypertext, in a new pane.  Information that is to be
;;; hotlinked is printed using HYPERPRINT, as described below.

;;; (hyperprint <thing>) displays the object <thing> on the current hypertext 
;;; pane as something with a hot link.  

;;; What happens when a hotlink is selected is determined by the generic 
;;; procedure QUESTIONS-FOR.  (If this looks familiar, it should -- I'm trying 
;;; to standardize this hypertext system so that we can have all of our systems
;;; loaded up at the same time as needed without interfering with each other!)
;;; It is recommended that, whenever possible, system-specific datatypes are 
;;; used for hotlinks, since there is likely to be contention in methods for 
;;; built-in types such as LIST or SYMBOL.

;;; (QUESTIONS-FOR <object> <context>) provides a list of conses representing 
;;; the questions and/or commands appropriate for that object in that context.
;;; Examples of contexts include designs in AVLs and LTREs and GIZMOs in GIZMO.
;;; The CAR of each cons is a string suitable for displaying in a menu, and the 
;;; CDR is a symbol which is recognized internally as denoting that question or 
;;; command.

;;; (FIND-ANSWER <object> <question-type> <context>) supplies the method to 
;;; carry out the work indicated by the result of QUESTIONS-FOR.  

;;; The hypertext pane includes table of hotlinks associated with it.
;;; Each hotlink includes an object and a pointer back to the window it belongs to,
;;; and is a kind of hotspot.

(defparameter *hypertext-parent* *lisp-main-window*)
(defparameter *hypertext-window* nil)

(defparameter  +pale-yellow+ #.(make-rgb :red 255 :green 255 :blue 192))
(defparameter +medium-green+ #.(make-rgb :red 0 :green 128 :blue 0))

(defparameter *hypertext-default-color* +pale-yellow+)
(defparameter *hypertext-hotlink-color* +medium-green+)


;;; MACROS
;;; ----------------------------------------------------------------------------
(defmacro htshow (format-string &rest args)
   `(htformat (hypertext-pane *hypertext-window*)
       ,format-string ,@ args))

(defmacro as-hypertext (title context parent &rest body)
   `(let* ((*hypertext-window* (make-hypertext-window ,title ,context ,parent))
           (*standard-output* *hypertext-window*))
       ;; Lambda-bind both as a matter of defensive programming
       ;; always ensure that the explanation window is restored first
       (declare (special *hypertext-window*))
       ;;so that text does not obliterate background color
       (setf (background-mode *hypertext-window*) :TRANSPARENT)
       ,@ body))

(defmacro with-hypertext-parent (stream &rest forms)
   `(let ((*hypertext-parent* ,stream)) ,@ forms))

(defmacro with-hypertext-window-color (new-color &rest forms)
   `(let ((*hypertext-default-color* ,new-color))
       ,@ forms))

(defmacro with-drawing-options ((pane &key ink fill-texture paint-op line-width font)
                                &body body)
   (multiple-value-bind (let-list setup-forms cleanup-forms)
       (make-drawing-options-internals pane ink fill-texture paint-op line-width font)
      `(let* ,let-list
          (unwind-protect
               (progn
                  ,@setup-forms
                  ,@body)
             ,@cleanup-forms))))


;;; HYPERTEXT CLASS DEFINITIONS
;;; ----------------------------------------------------------------------------
(defclass hypertext-window (non-refreshing-window)
    ((questions :initform nil :initarg :questions :accessor questions)))

(defclass hypertext-pane (hotspot-mixin non-refreshing-pane)
    ((context
        :documentation "An object used to help evaluate objects"
        :type T
        :initform nil
        :initarg :context
        :accessor context)
     (contents
        :documentation "The strings displayed on it so far"
        :type vector 
        :initform (make-array 1 :fill-pointer 0)
        :accessor contents)))

(defmethod default-pane-class ((window hypertext-window)) 'hypertext-pane)

(defmethod hypertext-pane ((window hypertext-window)) 
   (frame-child window)) ;; Should be only one

(defmethod hypertext-pane ((w hypertext-pane)) w)

(defmethod hypertext-parent ((w hypertext-pane))
   (window-parent (window-parent w)))

(defmethod hypertext-parent ((w hypertext-window))
   (window-parent w))


;;; HOTLINKS
;;; ----------------------------------------------------------------------------
(defclass hotlink (hotspot)
    ((object
        :documentation "The object this hotlink is about"
        :type T ;; Fully general
        :initarg :object ;; Unchangeable
        :reader object)
     (pane
        :accessor pane 
        :initarg :pane
        :type basic-pane
        :initform nil)))


;;; LINSPECT, THE PRIMARY USER INTERFACE
;;; ----------------------------------------------------------------------------
(defmethod linspect ((object list) &optional (parent *hypertext-parent*))
   (linspect (get-tms-node object) parent))

(defmethod linspect ((object tms-node) &optional (parent *hypertext-parent*))
   (as-hypertext 
      (format nil "Inspecting ~A" (view-node object))
      (tms-node-ltms object) ;;context
      parent
      (let* ((form (datum-lisp-form (tms-node-datum object)))
             (asns (assumptions-of form))
             (conseqs (consequences-of form)))
         (do-hyper-assumptions object)
         (do-hyper-why object)
         (do-hyper-consequences object))))


;;; INTERFACE TO USER-SUPPLIED QUESTIONS FOR AN OBJECT
;;; ----------------------------------------------------------------------------
(defun access-hypertext-for (object hypertext-pane)
   (let* ((questions-alist (questions-for object (context hypertext-pane)))
          (choice (ask-hypertext-query  
                   questions-alist hypertext-pane)))
      (when choice
         ;; Assume it should always appear in the same parent window
         (with-hypertext-parent (hypertext-parent hypertext-pane)
            (with-hourglass
               (find-answer object choice (context hypertext-pane)))))))

(defun ask-hypertext-query (questions-alist pane)
   "Uses popup menu"
   (let* ((menu (open-menu questions-alist 'pop-up-menu
                   (window-parent pane)
                   :name :hypertext-query-menu))
          (choice (pop-up-menu menu)))
      (close menu)
      choice))


;;; HYPERTEXT QUESTIONS AND ANSWERS
;;; ----------------------------------------------------------------------------
(defmethod questions-for ((node tms-node) (context LTMS))
   (nconc (if (known-node? node) 
             `(,(cons "Inspect" :INSPECT)
               ,(cons "Why?" :WHY)
               ,(cons "Consequences"  :CONSEQUENCES)
               ,(cons "Assumptions"  :ASSUMPTIONS)))
      (if (enabled-assumption? node)
         (list (cons "Retract"  :RETRACT)))
      (if (unknown-node? node)
         (list (cons "Assume true"  :ASSUME-TRUE)
            (cons "Assume false" :ASSUME-FALSE)
            (cons "Assert true"  :ASSERT-TRUE)
            (cons "Assert false" :ASSERT-FALSE)))))

(defmethod find-answer ((node tms-node)
                        (question-type (eql :INSPECT))
                        (context LTMS))
   (linspect node))

(defmethod find-answer ((node tms-node)
                        (question-type (eql :WHY))
                        (context LTMS))
   (hyper-why node))

(defmethod find-answer ((node tms-node)
                        (question-type (eql :ASSUMPTIONS))
                        (context LTMS))
   (hyper-assumptions node))

(defmethod find-answer ((node tms-node)
                        (question-type (eql :CONSEQUENCES))
                        (context LTMS))
   (hyper-consequences node))

(defmethod find-answer ((node tms-node)
                        (question-type (eql :RETRACT))
                        (context LTMS))
   (gui-retract node))

(defmethod find-answer ((node tms-node)
                        (question-type (eql :ASSUME-TRUE))
                        (context LTMS))
   (gui-assume node :TRUE))

(defmethod find-answer ((node tms-node)
                        (question-type (eql :ASSUME-FALSE))
                        (context LTMS))
   (gui-assume node :FALSE))

(defmethod find-answer ((node tms-node)
                        (question-type (eql :ASSERT-TRUE))
                        (context LTMS))
   (if (yes-or-no-p "Assertions cannot be retracted.  Are you sure?")
      (gui-assert node :TRUE)))

(defmethod find-answer ((node tms-node)
                        (question-type (eql :ASSERT-FALSE))
                        (context LTMS))
   (if (yes-or-no-p "Assertions cannot be retracted.  Are you sure?")
      (gui-assert node :FALSE)))


;;; HYPERTEXT MACHINERY FOR RESPONDING TO QUESTIONS AND ANSWERS
;;; ----------------------------------------------------------------------------
(defun hyper-why (node &optional (parent *hypertext-parent*))
   (as-hypertext 
      (format nil "Why ~A?" (node-string node))
      (tms-node-ltms node)
      parent
      (do-hyper-why node)))

(defun do-hyper-why (node)
    (let ((form (datum-lisp-form (tms-node-datum node))))
       (cond ((unknown-node? node)
              (htshow "~&It isn't known."))
             ((eq :ENABLED-ASSUMPTION (tms-node-support node))
              (htshow "~&It is an assumption.~%The source is ~A."
                 (datum-assumption? (tms-node-datum node))))
             ((premise? form)
              (htshow "~&It is a premise.~%The source is ~A."
               (informant-of form)))
             (t (htshow "It is a consequence of~%~A on"
                   (informant-of form))
                (dolist (ante (antecedents-of form)(htshow "~%  "))
                   (htshow "~%   ")
                   (hyperprint (get-tms-node ante)))))))

(defun hyper-consequences (node &optional (parent *hypertext-parent*))
   (as-hypertext 
      (format nil "Consequences of ~A?" (node-string node))
      (tms-node-ltms node) ;;context
      parent
      (do-hyper-consequences node)))
      
(defun do-hyper-consequences (node)      
   (let* ((form (datum-lisp-form (tms-node-datum node)))
          (conseqs (consequences-of form)))
      (cond ((null conseqs)
             (htshow "~&No known consequences."))
            (t (htshow "~&There are ~D consequences:"
                  (length conseqs))
               (dolist (c conseqs (htshow "~%  "))
                  (htshow "~%  ")
                  (hyperprint (get-tms-node c)))))))

(defun hyper-assumptions (node &optional (parent *hypertext-parent*))
   (as-hypertext 
      (format nil "Assumptions underlying ~A?" (node-string node))
      (tms-node-ltms node)
      parent
      (do-hyper-assumptions node)))

(defun do-hyper-assumptions (node)      
   (let* ((form (datum-lisp-form (tms-node-datum node)))
          (asns (assumptions-of form)))
      (cond ((null asns)
             (htshow "~&No known assumptions."))
            (t (htshow "~&It rests on ~D assumptions:"
                  (length asns))
               (dolist (c asns (htshow "~%  "))
                  (htshow "~%  ")
                  (hyperprint (get-tms-node c)))))))


;;; ASSERTING, ASSUMING AND RETRACTING FACTS VIA INSPECTOR WINDOWS
;;; ----------------------------------------------------------------------------
(defun gui-retract (node)
   (let ((form (datum-lisp-form (tms-node-datum node))))
      (retract! form (informant-of form))))

(defvar *default-informant* :USER)

(defun gui-assume (node label)
   (let ((form (ecase label
                  (:TRUE (datum-lisp-form (tms-node-datum node)))
                  (:FALSE (list :NOT (datum-lisp-form (tms-node-datum 
                                                         node))))))) 
      (assume! form *default-informant*)))

(defun gui-assert (node label)
   (let ((form (ecase label
                  (:TRUE (datum-lisp-form (tms-node-datum node)))
                  (:FALSE (list :NOT (datum-lisp-form (tms-node-datum 
                                                         node)))))))
      (when (yes-or-no-p
               "Are you REALLY sure about asserting~%~A?" form) ;; Make pretty
         (assert! form *default-informant*))))


;;; HOTLINK MACHINERY
;;; ----------------------------------------------------------------------------
(defmethod add-hotlink ((pane hypertext-pane) ;; 
                        (phrase string) ;; String corresponding to object
                        (expl T)) ;; The object itself
   ;; First cache the current position of the window, which will be the upper left
   ;; corner of the hotspot.
   (let ((*hypertext-hotlink-color* (link-color-for expl)))
      (with-drawing-options (pane :ink *hypertext-hotlink-color*)
         (let ((hotlink (make-instance 'hotlink :object expl
                           :highlight-style :outline
                           :pane pane)))
            (nmake-box-relative (hotspot-box hotlink)
               (position-x (current-position pane))
               (position-y (current-position pane))
               (stream-string-width pane phrase)
               (line-height pane))
         (add-hotspot pane hotlink) ;; Let window system know about the hotspot
            (htformat pane "~A" phrase) ;; And print it
            hotlink))))

(defun link-color-for (node)
   (cond ((tms-node-assumption? node)
          +medium-green+)
         ((known-node? node)
          blue)
         (t black)))


;;; EVENTS FOR HOTLINKS
;;; ----------------------------------------------------------------------------
(defconstant xor 195) 

(defmethod event ((hotlink hotlink)
                  (event (eql mouse-left-down)) buttons data time)
   (unless (dead-hotlink? hotlink)
      (invert-hotspot hotlink)
      (access-hypertext-for (object hotlink)
         (pane hotlink))
      (invert-hotspot hotlink)))

(defmethod dead-hotlink? ((h hotlink))
   (or (null (pane h))
       (closed-stream-p (pane h))))
   
(defmethod invert-hotspot ((self hotlink) &aux pane)
   "Don't destroy background with highlighting of hotspots"
   (when (setq pane (pane self))
      (with-drawing-options (pane :ink *hypertext-hotlink-color* 
                               :paint-op xor :line-width 1)
         (draw-box pane (hotspot-box self)))))


;;; HTFORMAT HANDLES THE PRODUCTION OF REFRESHABLE STRINGS
;;; ----------------------------------------------------------------------------
(defmethod htformat ((stream hypertext-window)
                     (format-string string)
                     &rest args)
   (error "Should be calling pane not window:~%~A, ~A, ~A"
      stream format-string args))

(defmethod htformat ((stream hypertext-pane)
                     (format-string string)
                     &rest args)
   (select-window (window-parent stream) t)
   ;; Create the string
   (let ((the-string (format nil "~1{~:}" format-string args))
         (color (foreground-color stream)))
      ;; Glue new stuff onto the contents, at the end.
      (vector-push-extend (list color the-string) (contents stream))
      ;; Spit it out
      (with-drawing-options (stream :ink color)
         (princ the-string stream))))


;;; CLEARING AND REFRESHING
;;; ----------------------------------------------------------------------------
;;; Because the hypertext pane is non-refreshing, we handle this ourselves.

(defmethod clear-hypertext ((window hypertext-window))
   (unless (closed-stream-p window)
      (clear-hypertext (hypertext-pane window))))

(defmethod clear-hypertext ((pane hypertext-pane))
   ;; This method ensures hotlinks are cleared as well as
   ;; just clearing the page.
   (dolist (hotlink (hotspots pane))
      ;; Prevent dangling event handlers
      (remove-hotspot pane hotlink))
   (setf (fill-pointer (contents pane)) 0) ;; Clear cached contents
   (clear-page pane)
   (move-to-x-y pane 0 0))

(defmethod redisplay-window ((pane hypertext-pane)
                             &optional box &aux old-color)
   (unless (closed-stream-p pane)
      (without-interrupts    
         (acl:clear-page pane)
         (setf old-color (foreground-color pane))
         (let ((contents (contents pane)))
            (dotimes (i (fill-pointer contents))
               (dolist (item (aref contents i))
                  (if (stringp item)
                     (princ item pane)
                     (set-foreground-color pane item)))))
         (set-foreground-color pane old-color))))
   

;;; CREATING HYPERTEXT WINDOWS
;;; ----------------------------------------------------------------------------
(defun make-hypertext-window (title context &optional (parent *lisp-main-window*))
   (open-stream 'hypertext-window
      parent
      :IO
      :name title
      :context context
      :title title
      :background-color *hypertext-default-color*
      :window-border nil
      :left-attachment nil :right-attachment nil
      :top-attachment nil :bottom-attachment nil
      :user-movable t
      :user-resizable t
      :user-closable T 
      :user-shrinkable nil
      :user-scrollable t
      :font #.(make-font :modern :courier\ new 16 nil)
      :pop-up-p nil
      :window-interior (hypertext-default-size)))


;;; DEFAULT INSPECTOR WINDOW SIZE AND PLACEMENT
;;; ----------------------------------------------------------------------------
(defparameter *inspector-width%* 35)
(defparameter *inspector-height%* 20)

(defun hypertext-default-size ()
   (make-box-relative 
      (random-start-left) (random-start-top)
      (round (* (/ *inspector-width%* 100) (right-margin *screen* t)))
      (round (* (/ *inspector-height%* 100) (bottom-margin *screen* t)))))
      
(defun random-start-left ()
   "Choose a random starting point so that windows don't appear in same place"
   (let ((screen-fraction (round (right-margin *screen* t) 10)))
      (+ screen-fraction (random screen-fraction))))
      
(defun random-start-top ()
   "Choose a random starting point so that windows don't appear in same place"
   (let ((screen-fraction (round (bottom-margin *screen* t) 10)))
      (+ screen-fraction (random screen-fraction))))


;;; CLOSING ALL LTRE INSPECTOR WINDOWS
;;; ----------------------------------------------------------------------------
(defun close-linspectors (&rest args)
   "Close all LTRE inspector windows"
   (declare (ignore args))
   (dolist (window (windows *lisp-main-window*))
      (when (and (eql (type-of window) 'hypertext-window)
                 (not (closed-stream-p window)))
         (close window))))

;; Add close command to Tools menu
(when (and (member :ACLPC *features*)
           (not (member :RUNTIME-SYSTEM *features*))
           (not (find-named-object :close-linspectors top::*tools-menu*)))
   (add-to-menu top::*tools-menu* 
      (make-menu-item
         :title "Close LTRE Inspectors"  
         :name :close-linspectors
         :value 'close-linspectors)))


;;; AUXILIARY FUNCTIONS
;;; ----------------------------------------------------------------------------
(defmethod hyperprint ((arg T))
   "Called within AS-HYPERTEXT, produces hotlink"
   ;; The really generic thing -- most datatypes will
   ;; want to do something more subtle
   (add-hotlink (hypertext-pane *hypertext-window*)
    (format nil "~A" arg)
    arg))

(defmethod hyperprint ((node tms-node))
   (hyperprint-helper node
    (concatenate 'string
       (symbol-name (tms-node-label node)) " "
       (node-string node))))

(defun hyperprint-helper (object string)
   "For defining new system-specific methods for hyperprint"
   (add-hotlink (hypertext-pane *hypertext-window*)
      string object))

(defun make-drawing-options-internals (pane ink fil-tex paint-op line-width font)
   (let* ((pane-sym (gensym "PANE"))
          (ink-sym (when ink (gensym "INK")))
          (fil-tex-sym (when fil-tex (gensym "FILL-TEXTURE")))
          (p-op-sym (when paint-op (gensym "PAINT-OPERATION")))
          (lw-sym (when line-width (gensym "LINE-WIDTH")))
          (font-sym (when font (gensym "FONT"))))
      (values 
         `((,pane-sym ,pane)
           ,@(when ink `((,ink-sym (foreground-color ,pane-sym))))
           ,@(when fil-tex `((,fil-tex-sym (fill-texture ,pane-sym))))
           ,@(when paint-op `((,p-op-sym (paint-operation ,pane-sym))))
           ,@(when line-width `((,lw-sym (line-width ,pane-sym))))
           ,@(when font `((,font-sym (font-handle ,pane-sym)))))
         `(,@(when ink `((set-foreground-color ,pane-sym ,ink)))
           ,@(when fil-tex `((set-fill-texture ,pane-sym ,fil-tex)))
           ,@(when paint-op `((setf (paint-operation ,pane-sym) ,paint-op)))
           ,@(when line-width `((set-line-width ,pane-sym ,line-width)))
           ,@(when font `((set-font ,pane-sym ,font))))
         `(,@(when ink `((set-foreground-color ,pane-sym ,ink-sym)))
           ,@(when fil-tex `((set-fill-texture ,pane-sym ,fil-tex-sym)))
           ,@(when paint-op `((setf (paint-operation ,pane-sym) ,p-op-sym)))
           ,@(when line-width `((set-line-width ,pane-sym ,lw-sym)))
           ,@(when font `((set-font ,pane-sym ,font-sym)))))))
;;; ----------------------------------------------------------------------------
;;; END OF CODE