;;;; -*-  Mode: LISP; Syntax: Common-Lisp; Base: 10                          -*-
;;;; ---------------------------------------------------------------------------
;;;; File name: runrules.lsp
;;;;    System: LTRE
;;;;   Version: 4.1
;;;;    Author: Kenneth D. Forbus, Johan de Kleer, John Everett
;;;;   Created: Jan 1, 1996
;;;;  Modified: everett on Sat Jul 6 20:44:59 1996
;;;;   Purpose: Rule running and enqueuing procedures for GC version of LTRE
;;;; Copyright: 1986, 1989, 1990, 1991, 1995 Kenneth D. Forbus, 
;;;;            Northwestern University, and Johan de Kleer, Xerox Corporation.
;;;;            All Rights Reserved.
;;;; ---------------------------------------------------------------------------

(in-package :LTRE)


;;; GLOBALS
;;; ----------------------------------------------------------------------------
(defvar *current-env* nil "Bindings that the current rule will be applied to")


;;; INSERTING NEW RULES
;;; ----------------------------------------------------------------------------
;;; This function inserts a rule into the database.  Rule forms compile into
;;; calls to this function, so it is quite active during runtime.

#+DEBUG-LTRE-RULES
(defun insert-rule (dbclass name matcher body)
   (let ((*LTRE* (dbclass-ltre dbclass))
         (*current-rule*
            (make-rule
               :DBCLASS dbclass
               :NAME name  ;;useful for debugging, but consumes space
               :MATCHER matcher
               :BODY body
               :COUNTER (incf (ltre-rule-counter *LTRE*)))))
      (push *current-rule* (dbclass-rules dbclass))
      (dolist (candidate (dbclass-facts dbclass))
         (try-rule-on *current-rule* candidate))))

#-DEBUG-LTRE-RULES
(defun insert-rule (dbclass matcher body)
   (let ((*LTRE* (dbclass-ltre dbclass))
         (*current-rule*
            (make-rule
               :DBCLASS dbclass
               :MATCHER matcher
               :BODY body
               :COUNTER (incf (ltre-rule-counter *LTRE*)))))
      (push *current-rule* (dbclass-rules dbclass))
      (dolist (candidate (dbclass-facts dbclass))
         (try-rule-on *current-rule* candidate))))

;;; MATCHING RULES
;;; ----------------------------------------------------------------------------
(defun try-rules (datum)
   (dolist (*current-rule* (dbclass-rules (datum-dbclass datum)))
      (dbg-ltre :TRY-RULES "~&Trying ~A on ~A~%"
         *current-rule* (view-node (datum-tms-node datum)))
      (try-rule-on *current-rule* datum)))

(defun try-rule-on (rule datum)
   (let ((*LTRE* (datum-ltre datum))
         (lisp-form (datum-lisp-form datum)))
      (multiple-value-bind (okay? bindings outer-triggers)
          (funcall (rule-matcher rule) lisp-form)
         (when okay?
            (let ((node (datum-tms-node datum)))
               (push (cons node bindings) (rule-bindings rule))
               (when (funcall (ltre-gc-test *ltre*) lisp-form)
                  ;;;Used to remove rule from node's rule cache
                  (push (cons node outer-triggers)
                     (rule-triggers rule))
                  ;;(push node (rule-tms-nodes rule))
                  (push rule (tms-node-rules node))))
            (enqueue rule *LTRE*)))))


;;; CACHING PARTIAL MATCHES FOR FUTURE RESCHEDULING
;;; ----------------------------------------------------------------------------
;;; Adding n cond clauses, where n is the number of triggers, to each rule body 
;;; results in significant rule code bloat.  Instead, we call a procedure that 
;;; caches the rule on the right node.

(defun cache-rule-on-node (node label)
   "Called at runtime to cache partially executed rules on correct node"
   (dbg-ltre :node-cache "Caching env on ~A ~A~%" node (tms-node-label node))
   (ecase label
      (:TRUE
       (push *current-env* (tms-node-true-rule-bindings node))
       (push *current-rule* (tms-node-true-rules node)))
      (:FALSE
       (push *current-env* (tms-node-false-rule-bindings node))
       (push *current-rule* (tms-node-false-rules node)))))


;;; QUEUEING THE RULES FOR EXECUTION
;;; ----------------------------------------------------------------------------
;;; Each rule instance is only executed once per set of bindings, but a 
;;; rule-struct may contain a rule that is executed once on several different 
;;; sets of bindings.  Since threading threaded rules shreds the thread, we 
;;; first check in enqueue to see if the rule is threaded, and only if it isn't 
;;; do we add it to the queue.  However, the bindings for that rule instance 
;;; are pushed onto the rule-struct's rule-bindings slot.  Similarly, dequeue 
;;; only pops the queue when exactly one set of bindings remains to be used.

(defun enqueue (rule &optional (*ltre* *ltre*) &aux q)
   "Only add new rules to queue--rules already there get more bindings"
   (setq q (ltre-rule-queue *ltre*))
   (dbg-ltre :rule-q "Queueing up ~A with ~A bindings"
      rule (length (rule-bindings rule)))
   (unless (rule-p q) ;;we already know it's dirty
      (setf (ltre-rule-thread-dirty? *ltre*) T))
   (unless (rule-thread rule)
      (setf (rule-thread rule) q)
      (setf (ltre-rule-queue *ltre*) rule)))

;;; Rules are popped off queue via shiftf, which efficiently performs multiple 
;;; setqs.  In (shiftf a b nil), a is setq'd to b, and b is setq'd to nil
;;; Because a rule on the queue may have bindings deleted by a gc operation, we 
;;; only return a rule if it has at least one set of bindings.

(defun dequeue (*ltre*)
   "Return next rule on q and pop q if rule has one or no bindings"
   (let ((rule (ltre-rule-queue *ltre*)))
      (when (rule-p rule)
         (let ((bindings (rule-bindings rule)))
            (when (null (cdr bindings)) ;;1 or 0 bindings left
               (dbg-ltre :RULE-Q "Popping ~A with ~A bindings"
                  rule (length bindings))
               (shiftf (ltre-rule-queue *ltre*) (rule-thread rule) nil))
            (when bindings
               (dbg-ltre :RULE-Q "Dequeue returning ~A" rule)
               rule)))))
   
(defun enqueue-cached-true-rules (node)
   "Called by set-truth to queue up rules cached on node"
   (do ((rule (pop (tms-node-true-rules node))
              (pop (tms-node-true-rules node)))
        (env (pop (tms-node-true-rule-bindings node))
             (pop (tms-node-true-rule-bindings node))))
       ((null rule))
      (push env (rule-bindings rule))
      (enqueue rule)))

(defun enqueue-cached-false-rules (node)
   "Called by set-truth to queue up rules cached on node"
   (do ((rule (pop (tms-node-false-rules node))
              (pop (tms-node-false-rules node)))
        (env (pop (tms-node-false-rule-bindings node))
             (pop (tms-node-false-rule-bindings node))))
       ((null rule))
      (push env (rule-bindings rule))
      (enqueue rule)))


;;; EXECUTING RULES THAT HAVE MATCHED
;;; ----------------------------------------------------------------------------
(defun run-rules (&optional (*LTRE* *LTRE*))
   (do ((*current-rule* (dequeue *LTRE*) (dequeue *LTRE*))
        (counter 0))
       ((null *current-rule*)
        (dbg-ltre :RUN-RULES "  ~A rules run" counter)
        (incf (ltre-rules-run *LTRE*) counter)
        counter)
      (let ((*current-env* (pop (rule-bindings *current-rule*))))
         (dbg-ltre :run-rules "~&RUN-RULES: CURRENT-ENV: ~A~%" *current-env*)
         (dbg-ltre :run-rules "~&Applying ~A to ~A~%"
            *current-rule* *current-env*)
         (apply (rule-body *current-rule*) *current-env*)
         (incf counter))))

(defun run-one-rule (&optional (*LTRE* *LTRE*))
   "Runs one rule if possible, and returns T if more rules are on queue"
   (let ((*current-rule* (dequeue *LTRE*)))
      (when *current-rule*
         (let ((*current-env* (pop (rule-bindings *current-rule*))))
            (dbg-ltre :RUN-RULES "Executing single rule ~A" *current-rule*)
            (incf (ltre-rules-run *LTRE*))
            (apply (rule-body *current-rule*) *current-env*))
         (rules-waiting? *LTRE*))))

(defun rules-waiting? (&optional (*LTRE* *LTRE*))
   (rule-p (ltre-rule-queue *ltre*)))


;;; GARBAGE COLLECTING BINDINGS ON RULES AND DELETING RULES
;;; ----------------------------------------------------------------------------
;;; Tricky part here is if rule-bindings become nil while on q, the dequeue 
;;; procedure has to throw away the rule, because it will choke the apply in 
;;; run-rules.

(defun delete-bindings (node rules)
   (dolist (rule rules)
      (when (rule-thread rule)
         (setf (rule-bindings rule)
               (delete-if #'(lambda (b) (eq node (car b)))
                  (rule-bindings rule)))
         ;;Remove trigger sets containing node from rule
         (setf (rule-triggers rule)
               (delete-if #'(lambda (triggers) (member node triggers))
                  (rule-triggers rule))))))

(defun delete-rules (rules)
   (mapc 'delete-rule rules))

(defun delete-rule (rule &aux rule-class)
   (setf (rule-thread rule) nil)
   ;;Using the list of triggers in place of the rule-tms-nodes slot
   (dolist (triggers (rule-triggers rule))
      (setf (tms-node-rules (car triggers))
            (delete rule (tms-node-rules (car triggers)))))
   (setf (rule-tms-nodes rule) nil)
   (setf (rule-triggers rule) nil)
   (setq rule-class (rule-dbclass rule))
   (setf (dbclass-rules rule-class)
         (delete rule (dbclass-rules rule-class) :test #'eq)))

(defun delete-node-from-rule-node-cache (rule node)
   "Called by delete-node to ensure all pointers to node go away"
   (setf (rule-tms-nodes rule)
         (delete node (rule-tms-nodes rule) :test #'eq))
   (setf (rule-triggers rule)
         (delete-if #'(lambda (triggers)
                         (member node triggers :test #'eq))
            (rule-triggers rule))))


;;; DISPLAY ROUTINES
;;; ----------------------------------------------------------------------------
(defun show-rules (&optional (*LTRE* *LTRE*) (stream *standard-output*)
			     &aux counter)
  (setq counter 0)
  (format t "~%The rules in ~A are:" (ltre-title *LTRE*))
  (map-dbclass
   #'(lambda (dbclass)
        (format *debug-io* "~%DBCLASS ~A------------" (dbclass-name dbclass))
        (dolist (rule (sort (copy-list (dbclass-rules dbclass))
                         #'< :key #'rule-counter))
           (incf counter)
           (print-rule rule stream)
           ;;;(dbg-show-rule rule stream)
           )))
    counter)

(defun print-rule (rule &optional (stream *standard-output*))
  (format stream "~%<Rule ~A>" (rule-counter rule)))

(defun test-rule-expansion ()
 (pprint (macroexpand
	  '(rule ((:TRUE (foo ?p ?q) :VAR ?f1)
		  (:FALSE (bar ?q ?r) :VAR ?f2)
		  (:INTERN (interesting ?r) :VAR ?f3))
		 (rassert! (:TAXONOMY ?p ?q ?f3) (:RANDOM-TEST ?f1 ?f2))))))
 
(defun get-rule (num &optional (*LTRE* *LTRE*))
   (map-dbclass
      #'(lambda (dbclass)
           (dolist (rule (dbclass-rules dbclass))
           (when (= (rule-counter rule) num)
              (return-from GET-RULE rule))))))

;;; ----------------------------------------------------------------------------
;;; END OF CODE